/*******************************************************************************
findx - PLI application to find all the X values in a simulation
Chris Spear, 3/10/98

HOW TO USE findx.c:

With VCS 4.x on a Sun or HP, compile findx.c:
    cc -c findx.c -I$VCS_HOME/`vcs -platform`/lib

Compile and link model with:
    vcs model.v findx.o -P findx.tab

Call the application from either the model or the CLI with:
    $findx(scope [, verbose])

where scope is the highest scope name to be searched, and
verbose is an optional flag, set to 1 if you want more information:

    $findx(top);

*******************************************************************************/

/* Include files                                                              */
#include <stdio.h>	   /* C I/O routines			              */
#include "vcsuser.h"	   /* Base Verilog PLI definitions	              */
#include "acc_user.h"	   /* Access routine definitions	              */
#include "malloc.h"        /* Malloc madness                                  */
#include "string.h"        /* Better string definitions                       */


/* Arguments passed into a routine                                            */
#define ARG_RET 0
#define ARG1 1
#define ARG2 2
#define ARG3 3
#define ARG4 4

/* Size of the largest vector that can be read - could be as high as 
   1 million bits / (32 bits / word) */
#define MAX_VECTOR 10000

/* Global signal controlling if we are running in verbose mode               */
int fx_verbose = 0;

static int mask[32] = {0};



/*******************************************************************************
  Procedure - 
      findx(signal)

  Description - 
      Finds any X's on the given signal

  Arguments - 
      signal: acc handle to signal (net or reg) to test
*******************************************************************************/
void findx(signal)
    handle signal;
{
    int i, mask_idx,
      bit_size = acc_fetch_size(signal),
      lw_size = ((acc_fetch_size(signal) - 1) / 32) + 1;
    static s_acc_value value = {accBinStrVal};
    static int val_array[MAX_VECTOR];

    /* We all have our limitations.... */
    if (lw_size > MAX_VECTOR) {
      tf_message (ERR_ERROR, "FINDX", "TOOBIG",
		  "Vector too large, $findx limited to %d bits, change MAX_VECTOR and recompile",
		  MAX_VECTOR);
      tf_dofinish();
    }

    /* If this is the first time through, initialize value structure */
    if (value.format == accBinStrVal) {
      /* Get the value of the signal */
      value.format = accVectorVal;
      value.value.vector = (p_acc_vecval) val_array;
    }

    (void) acc_fetch_value (signal, "%%", &value);

    /* Look for those naughty bits */
    for (i = 0; i < lw_size; i++) {
	unsigned int abits = value.value.vector[i].aval;
	unsigned int bbits = value.value.vector[i].bval;
	
	if (fx_verbose) io_printf("%x/%x ", abits, bbits);

	/* A signal contains an X if both the a & b bits are 1 */
	/* Mask out the upper bits which are undefined         */
	if (bit_size > 31)
	  mask_idx = 31;
	else mask_idx = bit_size - 1;

	if (abits & bbits & mask[mask_idx]) {
	    char *vptr;
	    if (fx_verbose) io_printf("\n");
	    vptr = acc_fetch_value(signal, "%h", null);
	    io_printf("Signal %s has the value %s\n",
		      acc_fetch_fullname(signal), 
		      vptr);

	    /* Quit so we only print the message once for signals > 32 bits */
	    break;
	} /* if a&b */

	/* Another 32 bits done... */
	bit_size = bit_size - 32;

    } /* for i */

    if (fx_verbose) io_printf("\n");

    free (value.value.vector);

} /* findx */



/*******************************************************************************
  Procedure - 
      find_signals(scope_handle)

  Description - 
      Finds all the signals (nets & regs) in a given scope,
      then call findx to look for X's on those signals

  Arguments - 
      scope_handle: acc_handle to scope to search
*******************************************************************************/
void find_signals(scope_handle)
    handle scope_handle;
{
    handle obj_handle;
    static net_reg_list[3] = {accNet, accRegister, 0};

    /* Find all the nets and registers in this instance */
    obj_handle = NULL;
    while (obj_handle = acc_next(net_reg_list, scope_handle, obj_handle)) {
	if (fx_verbose) {
	    if (acc_fetch_type(obj_handle) == accNet) 
		io_printf("        Net");
	    else io_printf("        Reg");
	    io_printf(" %s = ",
		      acc_fetch_name(obj_handle));
	}
	findx(obj_handle);
    } /* while obj_handle */

} /* find_signals */



/*******************************************************************************
  Procedure - 
      find_modules(scope_handle)

  Description - 
      Recursively find all the modules below a given scope.
      For every module, call find_signals to look at the signals

  Arguments - 
      scope_handle: acc_handle to scope to search
*******************************************************************************/
void find_modules(mod_handle)
    handle mod_handle;
{
    handle inst_handle;
    static int inst_list[2] = {accModuleInstance, 0};

    if (fx_verbose)
	io_printf("module %s\n",
		  acc_fetch_name(mod_handle));

    /* Check the signals in the current module */
    find_signals(mod_handle);

    /* Step through the module instances */
    inst_handle = NULL;
    while (inst_handle = acc_next(inst_list, mod_handle, inst_handle)) {
	if (fx_verbose)
	    io_printf("instance %s ", 
		      acc_fetch_name(inst_handle));

	find_modules(inst_handle);
    } /* while mod_handle */

} /* find_modules */




/*******************************************************************************
  $findx routines
  $findx(scope [, verbose]);

  Procedure -
     findx_call

  Description -
     Top procedure for $findx task, looks through all modules and
     instances and calls find_signals

  Arguments -
     None directly.
     tf_arg 1 - Top scope to be searched
     tf_arg 2 - verbose flag [optional, default = quiet]
*******************************************************************************/
void findx_call()
{
    handle mod_handle, inst_handle;
    static int inst_list[2] = {accModuleInstance, 0};

    /* Reset environment for ACC routines */
    acc_initialize();
    
    /* Should we run in verbose mode? */
    if (tf_nump()) 
	fx_verbose = tf_getp(ARG2);

    if (fx_verbose)
	io_printf("Starting $findx\n");

    /* Get handle of the top module, passed as argument 1 */
    mod_handle = acc_handle_tfarg(1);

    /* Find all the sub-modules, recursively */
    find_modules(mod_handle);

    if (fx_verbose)
	io_printf("Finished $findx!\n");

} /* findx_call */



/*******************************************************************************
  Procedure -
     findx_check

  Description -
     Checks the arguments to the $findx application, could do more...

  Arguments -
     None directly
*******************************************************************************/
void findx_check()
{
    int nbytes;
    s_tfnodeinfo info;


    if (tf_nump() < 1) {
	tf_message (ERR_ERROR, "FindX", "FindX0",
		    "$findx has 1 required arguments - scope");
    }

    if (tf_nump() > 2) {
	tf_message (ERR_ERROR, "FindX", "FindX2",
		    "$findx has a max of 2 arguments - scope and verbose, you gave %d", tf_nump());
    }

    /* More checks should be put here... */


    /* Initialize array of mask bits */
    if (!mask[0]) {
      int i;
      for (i = 0; i<31; i++) {
	mask[i] = (2 << i) - 1;
	/* printf("%d: %0x\n", i, mask[i]); */
      } /* for i */
      mask[31] = -1;
      /* printf("31: %0x\n", mask[31]); */
    } /* if mask */


} /* findx_check */
